@@ -10,6 +10,8 @@
 #	pragma hdrstop
 #endif
 
+#define _CRT_SECURE_NO_WARNINGS 1
+
 const char gLibNcFTPVersion[] = kLibraryVersion;
 
 #ifdef NO_SIGNALS
@@ -135,6 +137,7 @@ CancelConnect(int signum)
 int
 OpenControlConnection(const FTPCIPtr cip, char *host, unsigned int port)
 {
+	struct sockaddr_in localAddr;
 	struct in_addr ip_address;
 	int err = 0;
 	int result;
@@ -143,7 +146,7 @@ OpenControlConnection(const FTPCIPtr cip, char *host, unsigned int port)
 	volatile int sock2fd = -1;
 	ResponsePtr rp = NULL;
 	char **volatile curaddr;
-	int hpok;
+	int hpok, hprc;
 	struct hostent hp;
 	char *volatile fhost;
 	unsigned int fport;
@@ -183,11 +186,19 @@ OpenControlConnection(const FTPCIPtr cip, char *host, unsigned int port)
 
 	cip->servCtlAddr.sin_port = (unsigned short) fport;
 
-	if (GetHostEntry(&hp, fhost, &ip_address, cip->buf, cip->bufSize) != 0) {
+	hprc = GetHostEntry(&hp, fhost, &ip_address, cip->buf, cip->bufSize);
+	if (hprc != 0) {
 		hpok = 0;
 		/* Okay, no Host entry, but maybe we have a numeric address
 		 * in ip_address we can try.
 		 */
+#ifdef DNSSEC_LOCAL_VALIDATION
+		if (hprc == -2) {
+			FTPLogError(cip, kDontPerror, "%s: untrusted DNS response.\n", fhost);
+			cip->errNo = kErrHostUnknown;
+			return (kErrHostUnknown);
+		}
+#endif
 		if (ip_address.s_addr == INADDR_NONE) {
 			FTPLogError(cip, kDontPerror, "%s: unknown host.\n", fhost);
 			cip->errNo = kErrHostUnknown;
@@ -219,6 +230,19 @@ OpenControlConnection(const FTPCIPtr cip, char *host, unsigned int port)
 			return (kErrNewStreamSocket);
 		}
 
+		/* On rare occasions, the user may specify a local IP address to use. */
+		if (cip->preferredLocalAddr.sin_family != 0) {
+			localAddr = cip->preferredLocalAddr;
+			localAddr.sin_port = 0;
+			if (BindToEphemeralPortNumber(sockfd, &localAddr, (int) cip->ephemLo, (int) cip->ephemHi) < 0) {
+				FTPLogError(cip, kDoPerror, "Could not bind the control socket");
+				result = kErrBindCtrlSocket;
+				cip->errNo = kErrBindCtrlSocket;
+				(void) SClose(sockfd, 3);
+				return (kErrBindCtrlSocket);
+			}
+		}
+
 		/* This doesn't do anything if you left these
 		 * at their defaults (zero).  Otherwise it
 		 * tries to set the buffer size to the
@@ -304,6 +328,19 @@ OpenControlConnection(const FTPCIPtr cip, char *host, unsigned int port)
 			 */
 			(void) memcpy(&cip->servCtlAddr.sin_addr, *curaddr, (size_t) hp.h_length);
 
+			/* On rare occasions, the user may specify a local IP address to use. */
+			if (cip->preferredLocalAddr.sin_family != 0) {
+				localAddr = cip->preferredLocalAddr;
+				localAddr.sin_port = 0;
+				if (BindToEphemeralPortNumber(sockfd, &localAddr, (int) cip->ephemLo, (int) cip->ephemHi) < 0) {
+					FTPLogError(cip, kDoPerror, "Could not bind the control socket");
+					result = kErrBindCtrlSocket;
+					cip->errNo = kErrBindCtrlSocket;
+					(void) SClose(sockfd, 3);
+					return (kErrBindCtrlSocket);
+				}
+			}
+
 			/* This doesn't do anything if you left these
 			 * at their defaults (zero).  Otherwise it
 			 * tries to set the buffer size to the
@@ -683,7 +720,7 @@ FTPSetStartOffset(const FTPCIPtr cip, longest_int restartPt)
 		if (result < 0) {
 			DoneWithResponse(cip, rp);
 			return (result);
-		} else if (result == 3) {
+		} else if (result <= 3) {
 			/* Success */
 			cip->hasREST = kCommandAvailable;
 		} else if (FTP_UNIMPLEMENTED_CMD(rp->code)) {
@@ -889,8 +926,8 @@ FTPFixClientDataAddr(const FTPCIPtr cip)
 
 
 
-static int
-BindToEphemeralPortNumber(int sockfd, struct sockaddr_in *addrp, int ephemLo, int ephemHi)
+int
+BindToEphemeralPortNumber(const int sockfd, struct sockaddr_in *const addrp, const int ephemLo, const int ephemHi)
 {
 	int i;
 	int result;
@@ -916,7 +953,8 @@ BindToEphemeralPortNumber(int sockfd, struct sockaddr_in *addrp, int ephemLo, in
 			result = bind(sockfd, (struct sockaddr *) addrp, sizeof(struct sockaddr_in));
 			if (result == 0)
 				break;
-			if ((errno != 999)
+			sleep(1);
+			if ((errno == 999)
 				/* This next line is just fodder to
 				 * shut the compiler up about variable
 				 * not being used.
@@ -939,6 +977,7 @@ OpenDataConnection(const FTPCIPtr cip, int mode)
 	int result;
 	int setsbufs;
 	size_t rbs, sbs;
+	int passiveAttemptsRemaining = cip->maxNumberOfSuccessivePASVAttempts;
 
 	/* Before we can transfer any data, and before we even ask the
 	 * remote server to start transferring via RETR/NLST/etc, we have
@@ -980,6 +1019,9 @@ tryPort2:
 
 	if ((cip->hasPASV == kCommandNotAvailable) || (mode == kSendPortMode)) {
 tryPort:
+		/* Use the same IP that we're using for our control connection,
+		 * which may have been set to something explicitly.
+		 */
 		cip->ourDataAddr = cip->ourCtlAddr;
 		cip->ourDataAddr.sin_family = AF_INET;
 
@@ -1018,108 +1060,129 @@ tryPort:
 		cip->dataPortMode = kSendPortMode;
 	} else {
 		/* Passive mode.  Let the other side decide where to send. */
-		
-		cip->servDataAddr = cip->servCtlAddr;
-		cip->servDataAddr.sin_family = AF_INET;
-		cip->ourDataAddr = cip->ourCtlAddr;
-		cip->ourDataAddr.sin_family = AF_INET;
+	
+		while (--passiveAttemptsRemaining >= 0) {
+			cip->servDataAddr = cip->servCtlAddr;
+			cip->servDataAddr.sin_family = AF_INET;
+			cip->ourDataAddr = cip->ourCtlAddr;
+			cip->ourDataAddr.sin_family = AF_INET;
 
-		if (FTPSendPassive(cip, &cip->servDataAddr, &weirdPort) < 0) {
-			FTPLogError(cip, kDontPerror, "Passive mode refused.\n");
-			cip->hasPASV = kCommandNotAvailable;
-			
-			/* We can try using regular PORT commands, which are required
-			 * by all FTP protocol compliant programs, if you said so.
-			 *
-			 * We don't do this automatically, because if your host
-			 * is running a firewall you (probably) do not want SendPort
-			 * FTP for security reasons.
-			 */
-			if (mode == kFallBackToSendPortMode)
-				goto tryPort;
-			result = kErrPassiveModeFailed;
-			cip->errNo = kErrPassiveModeFailed;
-			goto bad;
-		}
-		FTPFixServerDataAddr(cip);
-
-#ifdef HAVE_LIBSOCKS
-		cip->ourDataAddr.sin_port = 0;
-		if (Rbind(dataSocket, (struct sockaddr *) &cip->ourDataAddr,
-			(int) sizeof (cip->ourDataAddr),
-			cip->servCtlAddr.sin_addr.s_addr) < 0) 
-#else
-		if (BindToEphemeralPortNumber(dataSocket, &cip->ourDataAddr, (int) cip->ephemLo, (int) cip->ephemHi) < 0)
-#endif
-		{
-			FTPLogError(cip, kDoPerror, "Could not bind the data socket");
-			result = kErrBindDataSocket;
-			cip->errNo = kErrBindDataSocket;
-			goto bad;
-		}
-
-#ifdef NO_SIGNALS
-		result = SConnect(dataSocket, &cip->servDataAddr, (int) cip->connTimeout);
-#else	/* NO_SIGNALS */
-		if (cip->connTimeout > 0)
-			(void) alarm(cip->connTimeout);
-
-		result = connect(dataSocket, (struct sockaddr *) &cip->servDataAddr, (int) sizeof(cip->servDataAddr));
-		if (cip->connTimeout > 0)
-			(void) alarm(0);
-#endif	/* NO_SIGNALS */
-
-#ifdef NO_SIGNALS
-		if (result == kTimeoutErr) {
-			if (mode == kFallBackToSendPortMode) {
-				FTPLogError(cip, kDontPerror, "Data connection timed out.\n");
-				FTPLogError(cip, kDontPerror, "Falling back to PORT instead of PASV mode.\n");
-				(void) DisposeSocket(dataSocket);
+			if (FTPSendPassive(cip, &cip->servDataAddr, &weirdPort) < 0) {
+				FTPLogError(cip, kDontPerror, "Passive mode refused.\n");
 				cip->hasPASV = kCommandNotAvailable;
-				goto tryPort2;
-			}
-			FTPLogError(cip, kDontPerror, "Data connection timed out.\n");
-			result = kErrConnectDataSocket;
-			cip->errNo = kErrConnectDataSocket;
-		} else
-#endif	/* NO_SIGNALS */
-
-		if (result < 0) {
-#ifdef ECONNREFUSED
-			if ((weirdPort > 0) && (errno == ECONNREFUSED)) {
-#elif defined(WSAECONNREFUSED)
-			if ((weirdPort > 0) && (errno == WSAECONNREFUSED)) {
-#endif
-				FTPLogError(cip, kDontPerror, "Server sent back a bogus port number.\nI will fall back to PORT instead of PASV mode.\n");
-				if (mode == kFallBackToSendPortMode) {
-					(void) DisposeSocket(dataSocket);
-					cip->hasPASV = kCommandNotAvailable;
-					goto tryPort2;
-				}
-				result = kErrServerSentBogusPortNumber;
-				cip->errNo = kErrServerSentBogusPortNumber;
+				
+				/* We can try using regular PORT commands, which are required
+				 * by all FTP protocol compliant programs, if you said so.
+				 *
+				 * We don't do this automatically, because if your host
+				 * is running a firewall you (probably) do not want SendPort
+				 * FTP for security reasons.
+				 */
+				if (mode == kFallBackToSendPortMode)
+					goto tryPort;
+				result = kErrPassiveModeFailed;
+				cip->errNo = kErrPassiveModeFailed;
 				goto bad;
 			}
-			if (mode == kFallBackToSendPortMode) {
-				FTPLogError(cip, kDoPerror, "connect failed.\n");
-				FTPLogError(cip, kDontPerror, "Falling back to PORT instead of PASV mode.\n");
-				(void) DisposeSocket(dataSocket);
-				cip->hasPASV = kCommandNotAvailable;
-				goto tryPort2;
+			FTPFixServerDataAddr(cip);
+
+#ifdef HAVE_LIBSOCKS
+			cip->ourDataAddr.sin_port = 0;
+			if (Rbind(dataSocket, (struct sockaddr *) &cip->ourDataAddr,
+				(int) sizeof (cip->ourDataAddr),
+				cip->servCtlAddr.sin_addr.s_addr) < 0) 
+#else
+			if (BindToEphemeralPortNumber(dataSocket, &cip->ourDataAddr, (int) cip->ephemLo, (int) cip->ephemHi) < 0)
+#endif
+			{
+				FTPLogError(cip, kDoPerror, "Could not bind the data socket");
+				result = kErrBindDataSocket;
+				cip->errNo = kErrBindDataSocket;
+				goto bad;
 			}
-			FTPLogError(cip, kDoPerror, "connect failed.\n");
-			result = kErrConnectDataSocket;
-			cip->errNo = kErrConnectDataSocket;
-			goto bad;
+
+			result = SConnect(dataSocket, &cip->servDataAddr, (int) cip->connTimeout);
+
+			if (result == kTimeoutErr) {
+				if (mode == kFallBackToSendPortMode) {
+					FTPLogError(cip, kDontPerror, "Data connection timed out.\n");
+					if (passiveAttemptsRemaining == 0) {
+						FTPLogError(cip, kDontPerror, "Falling back to PORT instead of PASV mode.\n");
+						(void) DisposeSocket(dataSocket);
+						dataSocket = kClosedFileDescriptor;
+						if (cip->hasPASV == kCommandAvailabilityUnknown)
+							cip->hasPASV = kCommandNotAvailable;
+						goto tryPort2;
+					}
+				} else {
+					FTPLogError(cip, kDontPerror, "Data connection timed out.\n");
+				}
+				result = kErrConnectDataSocket;
+				cip->errNo = kErrConnectDataSocket;
+				if (passiveAttemptsRemaining == 0)
+					goto bad;
+			} else if (result < 0) {
+#ifdef ECONNREFUSED
+				if ((weirdPort > 0) && (errno == ECONNREFUSED))
+#elif defined(WSAECONNREFUSED)
+				if ((weirdPort > 0) && (errno == WSAECONNREFUSED))
+#endif
+				{
+					if (mode == kFallBackToSendPortMode) {
+						if (passiveAttemptsRemaining == 0) {
+							if (cip->hasPASV == kCommandAvailabilityUnknown)
+								cip->hasPASV = kCommandNotAvailable;
+							FTPLogError(cip, kDontPerror, "Server sent back a bogus port number.\nI will fall back to PORT instead of PASV mode.\n");
+							(void) DisposeSocket(dataSocket);
+							dataSocket = kClosedFileDescriptor;
+							goto tryPort2;
+						}
+					}
+					FTPLogError(cip, kDontPerror, "Server sent back a bogus port number.\n");
+					result = kErrServerSentBogusPortNumber;
+					cip->errNo = kErrServerSentBogusPortNumber;
+					if (passiveAttemptsRemaining == 0)
+						goto bad;
+				}
+				if ((mode == kFallBackToSendPortMode) && (passiveAttemptsRemaining == 0)) {
+					FTPLogError(cip, kDoPerror, "connect failed.\n");
+					FTPLogError(cip, kDontPerror, "Falling back to PORT instead of PASV mode.\n");
+					(void) DisposeSocket(dataSocket);
+					dataSocket = kClosedFileDescriptor;
+					if (cip->hasPASV == kCommandAvailabilityUnknown)
+						cip->hasPASV = kCommandNotAvailable;
+					goto tryPort2;
+				}
+				FTPLogError(cip, kDoPerror, "connect failed.\n");
+				result = kErrConnectDataSocket;
+				cip->errNo = kErrConnectDataSocket;
+			} else {
+				/* Success, break loop because we do not need to do PASV again. */
+				break;
+			}
+
+			if (dataSocket != kClosedFileDescriptor)
+				(void) DisposeSocket(dataSocket);
+
+			dataSocket = socket(AF_INET, SOCK_STREAM, 0);
+			if (dataSocket < 0) {
+				FTPLogError(cip, kDoPerror, "Could not get a data socket.\n");
+				result = kErrNewStreamSocket;
+				cip->errNo = kErrNewStreamSocket;
+				return result;
+			}
+			PrintF(cip, "Retrying PASV mode (%d %s left).\n", passiveAttemptsRemaining, (passiveAttemptsRemaining == 1) ? "try" : "tries");
 		}
-	
+
+		if (result != kNoErr)
+			goto bad;
+
 		/* Need to do this so we can figure out which port the system
 		 * gave to us.
 		 */
 		if ((result = GetSocketAddress(cip, dataSocket, &cip->ourDataAddr)) < 0)
 			goto bad;
 
-		cip->dataPortMode = kPassiveMode;
 		cip->hasPASV = kCommandAvailable;
 	}
 
