@@ -1,6 +1,6 @@
 /* io_put.c
  *
- * Copyright (c) 1996-2006 Mike Gleason, NcFTP Software.
+ * Copyright (c) 1996-2008 Mike Gleason, NcFTP Software.
  * All rights reserved.
  *
  */
@@ -10,6 +10,8 @@
 #	pragma hdrstop
 #endif
 
+#define _CRT_SECURE_NO_WARNINGS 1
+
 #ifndef O_BINARY
 	/* Needed for platforms using different EOLN sequence (i.e. DOS) */
 #	ifdef _O_BINARY
@@ -93,7 +95,7 @@ FTPPutBlock(
 	do {
 		if (! WaitForRemoteOutput(cip)) {	/* could set cancelXfer */
 			cip->errNo = result = kErrDataTimedOut;
-			FTPLogError(cip, kDontPerror, "Remote write timed out.\n");
+			FTPLogError(cip, kDontPerror, "Remote write timed out after " PRINTF_LONG_LONG " bytes had been sent.\n", cip->bytesTransferred);
 			return (result);
 		}
 		if (cip->cancelXfer > 0) {
@@ -106,16 +108,16 @@ FTPPutBlock(
 		if (nwrote < 0) {
 			if (nwrote == kTimeoutErr) {
 				cip->errNo = result = kErrDataTimedOut;
-				FTPLogError(cip, kDontPerror, "Remote write timed out.\n");
+				FTPLogError(cip, kDontPerror, "Remote write timed out after " PRINTF_LONG_LONG " bytes had been sent.\n", cip->bytesTransferred);
 			} else if (errno == EPIPE) {
 				cip->errNo = result = kErrSocketWriteFailed;
 				errno = EPIPE;
-				FTPLogError(cip, kDoPerror, "Lost data connection to remote host.\n");
+				FTPLogError(cip, kDoPerror, "Lost data connection to remote host after " PRINTF_LONG_LONG " bytes had been sent.\n", cip->bytesTransferred);
 			} else if (errno == EINTR) {
 				continue;
 			} else {
 				cip->errNo = result = kErrSocketWriteFailed;
-				FTPLogError(cip, kDoPerror, "Remote write failed.\n");
+				FTPLogError(cip, kDoPerror, "Remote write failed after " PRINTF_LONG_LONG " bytes had been sent.\n", cip->bytesTransferred);
 			}
 			(void) shutdown(cip->dataSocket, 2);
 			return (result);
