@@ -10,6 +10,43 @@
 #	pragma hdrstop
 #endif
 
+void
+FTPResetStatusVariables(const FTPCIPtr cip)
+{
+	cip->errNo = 0;
+	memset(&cip->lastFTPCmdResultStr, 0, sizeof(cip->lastFTPCmdResultStr));
+	cip->lastFTPCmdResultNum = 0;
+	cip->sec = 0;
+	cip->secLeft = 0;
+	cip->kBytesPerSec = 0;
+	cip->percentCompleted = 0;
+	cip->expectedSize = 0;
+	cip->mdtm = 0;
+	cip->nextProgressUpdate = 0;
+	cip->stalled = 0;
+	cip->dataTimedOut = 0;
+	cip->cancelXfer = 0;
+	cip->canceling = 0;
+	cip->canceled = 0;
+	cip->connected = 0;
+	cip->loggedIn = 0;
+	cip->bytesTransferred = 0;
+	cip->curTransferType = 0;
+	cip->startPoint = 0;
+	cip->dataSocketConnected = 0;
+	cip->totalDials = 0;
+	memset(&cip->connectTime, 0, sizeof(cip->connectTime));
+	memset(&cip->loginTime, 0, sizeof(cip->loginTime));
+	memset(&cip->disconnectTime, 0, sizeof(cip->disconnectTime));
+	memset(&cip->lastCmdStart, 0, sizeof(cip->lastCmdStart));
+	memset(&cip->lastCmdFinish, 0, sizeof(cip->lastCmdFinish));
+	cip->numDownloads = 0;
+	cip->numUploads = 0;
+	cip->numListings = 0;
+}	/* FTPResetStatusVariables */
+
+
+
 
 void
 FTPDeallocateHost(const FTPCIPtr cip)
@@ -35,6 +72,7 @@ FTPDeallocateHost(const FTPCIPtr cip)
 	DisposeSReadlineInfo(&cip->ctrlSrl);
 #endif
 	DisposeLineListContents(&cip->lastFTPCmdResultLL);
+
 }	/* FTPDeallocateHost */
 
 
@@ -393,6 +431,13 @@ FTPQueryFeatures(const FTPCIPtr cip)
 	if (strcmp(cip->magic, kLibraryMagic))
 		return (kErrBadMagic);
 
+	if (cip->serverType == kServerTypeMicrosoftFTP) {
+		cip->hasNLST_a = kCommandNotAvailable;
+		cip->hasNLST_d = kCommandNotAvailable;
+		cip->hasSITE_UTIME = kCommandNotAvailable;
+		cip->hasRETR_tar = kCommandNotAvailable;
+	}
+
 	if (cip->serverType == kServerTypeNetWareFTP) {
 		/* NetWare 5.00 server freaks out when
 		 * you give it a command it doesn't
@@ -409,6 +454,7 @@ FTPQueryFeatures(const FTPCIPtr cip)
 		cip->hasMLSD = kCommandNotAvailable;
 		cip->hasSITE_UTIME = kCommandNotAvailable;
 		cip->hasHELP_SITE = kCommandNotAvailable;
+		cip->hasRETR_tar = kCommandNotAvailable;
 		return (kNoErr);
 	}
 
@@ -427,122 +473,138 @@ FTPQueryFeatures(const FTPCIPtr cip)
 		cip->hasMLSD = kCommandNotAvailable;
 		cip->hasSITE_UTIME = kCommandNotAvailable;
 		cip->hasHELP_SITE = kCommandNotAvailable;
+		cip->hasMFMT = kCommandNotAvailable;
+		cip->hasMFF = kCommandNotAvailable;
+		cip->hasRETR_tar = kCommandNotAvailable;
+	}
+
+	/* We cheat here and pre-populate some
+	 * fields when the server is wu-ftpd.
+	 * This server is very common and we
+	 * know it has always had these.
+	 */
+	 if (cip->serverType == kServerTypeWuFTPd) {
+		cip->hasPASV = kCommandAvailable;
+		cip->hasSIZE = kCommandAvailable;
+		cip->hasMDTM = kCommandAvailable;
+		cip->hasMDTM_set = kCommandAvailable;
+		cip->hasREST = kCommandAvailable;
+		cip->NLSTfileParamWorks = kCommandAvailable;
+	} else if (cip->serverType == kServerTypeNcFTPd) {
+		cip->hasPASV = kCommandAvailable;
+		cip->hasSIZE = kCommandAvailable;
+		cip->hasMDTM = kCommandAvailable;
+		cip->hasREST = kCommandAvailable;
+		cip->NLSTfileParamWorks = kCommandAvailable;
+	}
+
+	if (cip->hasFEAT == kCommandNotAvailable) {
+		return (kNoErr);
 	}
 	
 	rp = InitResponse();
 	if (rp == NULL) {
 		cip->errNo = kErrMallocFailed;
 		result = cip->errNo;
-	} else if (cip->hasFEAT != kCommandNotAvailable) {
-		rp->printMode = (kResponseNoPrint|kResponseNoSave);
-		result = RCmd(cip, rp, "FEAT");
-		if (result < kNoErr) {
-			DoneWithResponse(cip, rp);
-			return (result);
-		} else if (result != 2) {
-			/* We cheat here and pre-populate some
-			 * fields when the server is wu-ftpd.
-			 * This server is very common and we
-			 * know it has always had these.
-			 */
-			 if (cip->serverType == kServerTypeWuFTPd) {
-				cip->hasPASV = kCommandAvailable;
-				cip->hasSIZE = kCommandAvailable;
-				cip->hasMDTM = kCommandAvailable;
-				cip->hasMDTM_set = kCommandAvailable;
-				cip->hasREST = kCommandAvailable;
-				cip->NLSTfileParamWorks = kCommandAvailable;
-			} else if (cip->serverType == kServerTypeNcFTPd) {
-				cip->hasPASV = kCommandAvailable;
-				cip->hasSIZE = kCommandAvailable;
-				cip->hasMDTM = kCommandAvailable;
-				cip->hasREST = kCommandAvailable;
-				cip->NLSTfileParamWorks = kCommandAvailable;
-			}
-
-			/* Newer commands are only shown in FEAT,
-			 * so we don't have to do the "try it,
-			 * then save that it didn't work" thing.
-			 */
-			cip->hasMLST = kCommandNotAvailable;
-			cip->hasMLSD = kCommandNotAvailable;
-		} else {
-			cip->hasFEAT = kCommandAvailable;
-
-			for (lp = rp->msg.first; lp != NULL; lp = lp->next) {
-				/* If first character was not a space it is
-				 * either:
-				 *
-				 * (a) The header line in the response;
-				 * (b) The trailer line in the response;
-				 * (c) A protocol violation.
-				 */
-				cp = lp->line;
-				if (*cp++ != ' ')
-					continue;
-				if (ISTRNCMP(cp, "PASV", 4) == 0) {
-					cip->hasPASV = kCommandAvailable;
-				} else if (ISTRNCMP(cp, "SIZE", 4) == 0) {
-					cip->hasSIZE = kCommandAvailable;
-				} else if (ISTRNCMP(cp, "MDTM", 4) == 0) {
-					cip->hasMDTM = kCommandAvailable;
-				} else if (ISTRNCMP(cp, "REST", 4) == 0) {
-					cip->hasREST = kCommandAvailable;
-				} else if (ISTRNCMP(cp, "UTIME", 5) == 0) {
-					cip->hasSITE_UTIME = kCommandAvailable;
-				} else if (ISTRNCMP(cp, "MLST", 4) == 0) {
-					cip->hasMLST = kCommandAvailable;
-					cip->hasMLSD = kCommandAvailable;
-					FTPExamineMlstFeatures(cip, cp + 5);
-				} else if (ISTRNCMP(cp, "CLNT", 4) == 0) {
-					cip->hasCLNT = kCommandAvailable;
-				} else if (ISTRNCMP(cp, "Compliance Level: ", 18) == 0) {
-					/* Probably only NcFTPd will ever implement this.
-					 * But we use it internally to differentiate
-					 * between different NcFTPd implementations of
-					 * IETF extensions.
-					 */
-					cip->ietfCompatLevel = atoi(cp + 18);
-				}
-			}
-		}
-
-		/* You can set cip->hasHELP_SITE to kCommandNotAvailable
-		 * if your host chokes (i.e. IBM Mainframes) when you
-		 * do "HELP SITE".
-		 */
-		ReInitResponse(cip, rp);
-		result = (cip->hasHELP_SITE == kCommandNotAvailable) ? (-1) : RCmd(cip, rp, "HELP SITE");
-		if (result == 2) {
-			cip->hasHELP_SITE = kCommandAvailable;
-			for (lp = rp->msg.first; lp != NULL; lp = lp->next) {
-				cp = lp->line;
-				if (strstr(cp, "RETRBUFSIZE") != NULL)
-					cip->hasSITE_RETRBUFSIZE = kCommandAvailable;
-				if (strstr(cp, "RBUFSZ") != NULL)
-					cip->hasSITE_RBUFSZ = kCommandAvailable;
-				/* See if RBUFSIZ matches (but not STORBUFSIZE) */
-				if (
-					((p = strstr(cp, "RBUFSIZ")) != NULL) &&
-					(
-					 	(p == cp) ||
-						((p > cp) && (!isupper((int) p[-1])))
-					)
-				)
-					cip->hasSITE_RBUFSIZ = kCommandAvailable;
-				if (strstr(cp, "STORBUFSIZE") != NULL)
-					cip->hasSITE_STORBUFSIZE = kCommandAvailable;
-				if (strstr(cp, "SBUFSIZ") != NULL)
-					cip->hasSITE_SBUFSIZ = kCommandAvailable;
-				if (strstr(cp, "SBUFSZ") != NULL)
-					cip->hasSITE_SBUFSZ = kCommandAvailable;
-				if (strstr(cp, "BUFSIZE") != NULL)
-					cip->hasSITE_BUFSIZE = kCommandAvailable;
-			}
-		}
-		DoneWithResponse(cip, rp);
+		return (result);
 	}
 
+	rp->printMode = (kResponseNoPrint|kResponseNoSave);
+	result = RCmd(cip, rp, "FEAT");
+	if (result < kNoErr) {
+		DoneWithResponse(cip, rp);
+		return (result);
+	} else if (result != 2) {
+		/* Newer commands are only shown in FEAT,
+		 * so we don't have to do the "try it,
+		 * then save that it didn't work" thing.
+		 */
+		cip->hasMLST = kCommandNotAvailable;
+		cip->hasMLSD = kCommandNotAvailable;
+		cip->hasMFMT = kCommandNotAvailable;
+		cip->hasMFF = kCommandNotAvailable;
+	} else {
+		cip->hasFEAT = kCommandAvailable;
+
+		for (lp = rp->msg.first; lp != NULL; lp = lp->next) {
+			/* If first character was not a space it is
+			 * either:
+			 *
+			 * (a) The header line in the response;
+			 * (b) The trailer line in the response;
+			 * (c) A protocol violation.
+			 */
+			cp = lp->line;
+			while ((*cp != '\0') && (isspace((int) *cp)))
+				cp++;
+			if (*cp == '\0')
+				continue;
+			if (ISTRNCMP(cp, "PASV", 4) == 0) {
+				cip->hasPASV = kCommandAvailable;
+			} else if (ISTRNCMP(cp, "SIZE", 4) == 0) {
+				cip->hasSIZE = kCommandAvailable;
+			} else if (ISTRNCMP(cp, "MDTM", 4) == 0) {
+				cip->hasMDTM = kCommandAvailable;
+			} else if (ISTRNCMP(cp, "MFMT", 4) == 0) {
+				cip->hasMFMT = kCommandAvailable;
+			} else if (ISTRNCMP(cp, "MFF", 3) == 0) {
+				cip->hasMFF = kCommandAvailable;
+			} else if (ISTRNCMP(cp, "REST", 4) == 0) {
+				cip->hasREST = kCommandAvailable;
+			} else if (ISTRNCMP(cp, "UTIME", 5) == 0) {
+				cip->hasSITE_UTIME = kCommandAvailable;
+			} else if (ISTRNCMP(cp, "MLST", 4) == 0) {
+				cip->hasMLST = kCommandAvailable;
+				cip->hasMLSD = kCommandAvailable;
+				FTPExamineMlstFeatures(cip, cp + 5);
+			} else if (ISTRNCMP(cp, "CLNT", 4) == 0) {
+				cip->hasCLNT = kCommandAvailable;
+			} else if (ISTRNCMP(cp, "Compliance Level: ", 18) == 0) {
+				/* Probably only NcFTPd will ever implement this.
+				 * But we use it internally to differentiate
+				 * between different NcFTPd implementations of
+				 * IETF extensions.
+				 */
+				cip->ietfCompatLevel = atoi(cp + 18);
+			}
+		}
+	}
+
+	/* You can set cip->hasHELP_SITE to kCommandNotAvailable
+	 * if your host chokes (i.e. IBM Mainframes) when you
+	 * do "HELP SITE".
+	 */
+	ReInitResponse(cip, rp);
+	result = (cip->hasHELP_SITE == kCommandNotAvailable) ? (-1) : RCmd(cip, rp, "HELP SITE");
+	if (result == 2) {
+		cip->hasHELP_SITE = kCommandAvailable;
+		for (lp = rp->msg.first; lp != NULL; lp = lp->next) {
+			cp = lp->line;
+			if (strstr(cp, "RETRBUFSIZE") != NULL)
+				cip->hasSITE_RETRBUFSIZE = kCommandAvailable;
+			if (strstr(cp, "RBUFSZ") != NULL)
+				cip->hasSITE_RBUFSZ = kCommandAvailable;
+			/* See if RBUFSIZ matches (but not STORBUFSIZE) */
+			if (
+				((p = strstr(cp, "RBUFSIZ")) != NULL) &&
+				(
+					(p == cp) ||
+					((p > cp) && (!isupper((int) p[-1])))
+				)
+			)
+				cip->hasSITE_RBUFSIZ = kCommandAvailable;
+			if (strstr(cp, "STORBUFSIZE") != NULL)
+				cip->hasSITE_STORBUFSIZE = kCommandAvailable;
+			if (strstr(cp, "SBUFSIZ") != NULL)
+				cip->hasSITE_SBUFSIZ = kCommandAvailable;
+			if (strstr(cp, "SBUFSZ") != NULL)
+				cip->hasSITE_SBUFSZ = kCommandAvailable;
+			if (strstr(cp, "BUFSIZE") != NULL)
+				cip->hasSITE_BUFSIZE = kCommandAvailable;
+		}
+	}
+	DoneWithResponse(cip, rp);
+
 	return (kNoErr);
 }	/* FTPQueryFeatures */
 
@@ -685,6 +747,49 @@ FTPInitialLogEntry(const FTPCIPtr cip)
 		}
 #endif
 
+#ifdef LINUX
+	{
+		FILE *etc_fp;
+		const char **etc_fn;
+		char etc_line[256], *etc_cp;
+
+		const char *etc_fnames[] = {
+			"/etc/yellowdog-release",
+			"/etc/gentoo-release",
+			"/etc/turbolinux-release",
+			"/etc/slackware-release",
+			"/etc/slackware-version",
+			"/etc/mandrake-release",
+			"/etc/debian_version",
+			"/etc/conectiva-release",
+			"/etc/SuSE-release",
+			"/etc/fedora-release",
+			"/etc/redhat-release",	/* Works for CentOS, too. */
+			"/etc/issue",
+			NULL,
+		};
+
+		for (etc_fn = etc_fnames; *etc_fn != NULL; etc_fn++) {
+			etc_fp = fopen(*etc_fn, "r");
+			if (etc_fp != NULL) {
+				PrintF(cip, "Contents of %.127s:\n", *etc_fn);
+				memset(etc_line, 0, sizeof(etc_line));
+				while (fgets(etc_line, sizeof(etc_line) - 1, etc_fp) != NULL) {
+					etc_cp = etc_line + strlen(etc_line) - 1;
+					while ((etc_cp >= etc_line) && (isspace((int) *etc_cp)))
+						etc_cp--;
+					++etc_cp;
+					*etc_cp = '\0';
+					if (etc_line[0] != '\0') {
+						PrintF(cip, "  %.127s\n", etc_line);
+					}
+				}
+				(void) fclose(etc_fp);
+			}
+		}
+	}
+#endif	/* LINUX */
+
 #if defined(HAVE_GNU_GET_LIBC_VERSION) && defined(HAVE_GNU_GET_LIBC_RELEASE)
 		PrintF(cip, "Glibc: %s (%s)\n",
 			gnu_get_libc_version(),
@@ -778,6 +883,7 @@ FTPOpenHost(const FTPCIPtr cip)
 		return (kErrNoHostSpecified);
 	}
 
+	FTPResetStatusVariables(cip);
 	FTPManualOverrideFeatures(cip);
 	FTPInitialLogEntry(cip);
 
@@ -909,12 +1015,15 @@ FTPInitConnectionInfo2(const FTPLIPtr lip, const FTPCIPtr cip, char *const buf,
 	cip->ctrlSocketR = kClosedFileDescriptor;
 	cip->ctrlSocketW = kClosedFileDescriptor;
 	cip->dataPortMode = kDefaultDataPortMode;
+	cip->maxNumberOfSuccessivePASVAttempts = kDefaultMaxNumberOfSuccessivePASVAttempts;
 	cip->dataSocket = kClosedFileDescriptor;
 	cip->lip = lip;
 	cip->hasPASV = kCommandAvailabilityUnknown;
 	cip->hasSIZE = kCommandAvailabilityUnknown;
 	cip->hasMDTM = kCommandAvailabilityUnknown;
 	cip->hasMDTM_set = kCommandAvailabilityUnknown;
+	cip->hasMFMT = kCommandAvailabilityUnknown;
+	cip->hasMFF = kCommandAvailabilityUnknown;
 	cip->hasREST = kCommandAvailabilityUnknown;
 	cip->hasNLST_a = kCommandAvailabilityUnknown;
 	cip->hasNLST_d = kCommandAvailabilityUnknown;
@@ -932,6 +1041,7 @@ FTPInitConnectionInfo2(const FTPLIPtr lip, const FTPCIPtr cip, char *const buf,
 	cip->hasSITE_SBUFSZ = kCommandAvailabilityUnknown;
 	cip->STATfileParamWorks = kCommandAvailabilityUnknown;
 	cip->NLSTfileParamWorks = kCommandAvailabilityUnknown;
+	cip->hasRETR_tar = kCommandAvailabilityUnknown;
 	cip->firewallType = kFirewallNotInUse;
 	cip->startingWorkingDirectory = NULL;
 	cip->shutdownUnusedSideOfSockets = 0;
