@@ -1,6 +1,6 @@
 /* util.c
  *
- * Copyright (c) 1992-2002 by Mike Gleason.
+ * Copyright (c) 1992-2006 by Mike Gleason.
  * All rights reserved.
  * 
  */
@@ -19,7 +19,7 @@ char gHome[256];
 char gShell[256];
 char gOurDirectoryPath[260];
 char gOurInstallationPath[260];
-int gNcFTP_Uses_Me_To_Quiet_Variable_Unused_Warnings = 0;
+const char *gNcFTP_Uses_Me_To_Quiet_Variable_Unused_Warnings = NULL;
 #ifdef ncftp
 static int gResolveSig;
 #endif
@@ -383,7 +383,7 @@ InitOurDirectory(void)
 			cp = strrchr(gOurDirectoryPath, '"');
 			if ((cp != NULL) && (cp[1] == '\0'))
 				*cp = '\0';
-		} else if ((cp = getenv("HOME")) != NULL) {
+		} else if ((cp = getenv("USERPROFILE")) != NULL) {
 			if (*cp == '"')
 				cp++;
 			(void) STRNCPY(gOurDirectoryPath, cp);
@@ -410,6 +410,8 @@ InitOurDirectory(void)
 
 #ifdef BINDIR
 	(void) STRNCPY(gOurInstallationPath, BINDIR);
+#elif defined(PREFIX_BINDIR)
+	(void) STRNCPY(gOurInstallationPath, PREFIX_BINDIR);
 #else
 	memset(gOurInstallationPath, 0, sizeof(gOurInstallationPath));
 #endif
@@ -509,15 +511,35 @@ InitUserInfo(void)
 
 
 
+char *
+TolowerStr(char *dst)
+{
+	if (dst == NULL)
+		return NULL;
+
+	while (*dst != '\0') {
+		if (isupper((int) *dst))
+			*dst = (char) tolower(*dst);
+		dst++;
+	}
+	return (dst);
+}	/* TolowerStr */
+
+
+
 
 int
-MayUseFirewall(const char *const hn, int firewallType, const char *const firewallExceptionList)
+MayUseFirewall(const char *const hn0, int firewallType, const char *const firewallExceptionList)
 {
 #ifdef HAVE_STRSTR
 	char buf[256];
 	char *tok;
 	char *parse;
 #endif /* HAVE_STRSTR */
+	char hn[80];
+
+	STRNCPY(hn, hn0);
+	TolowerStr(hn);
 
 	if (firewallType == kFirewallNotInUse)
 		return (0);
@@ -542,7 +564,7 @@ MayUseFirewall(const char *const hn, int firewallType, const char *const firewal
 		 */
 		(void) STRNCPY(buf, firewallExceptionList);
 		for (parse = buf; (tok = strtok(parse, ", \n\t\r")) != NULL; parse = NULL) {
-			if (strcmp(tok, "localdomain") == 0)
+			if (ISTREQ(tok, "localdomain"))
 				return (0);
 		}
 		/* fall through */
@@ -550,6 +572,7 @@ MayUseFirewall(const char *const hn, int firewallType, const char *const firewal
 
 #ifdef HAVE_STRSTR
 	(void) STRNCPY(buf, firewallExceptionList);
+	TolowerStr(buf);
 	for (parse = buf; (tok = strtok(parse, ", \n\t\r")) != NULL; parse = NULL) {
 		/* See if host or domain was from exclusion list
 		 * matches the host to open.
@@ -667,6 +690,9 @@ MyGetHostByName(char *const volatile dst, size_t dsize, const char *const hn, in
 #ifdef HAVE_INET_ATON
 	struct in_addr ina;
 #endif
+#ifdef DNSSEC_LOCAL_VALIDATION
+        val_status_t val_status;
+#endif
 
 #ifdef HAVE_INET_ATON
 	if (inet_aton(hn, &ina) != 0) {
@@ -705,12 +731,26 @@ MyGetHostByName(char *const volatile dst, size_t dsize, const char *const hn, in
 		osigalrm = NcSignal(SIGALRM, CancelGetHostByName);
 		if (t > 0)
 			(void) alarm((unsigned int) t);
+#ifndef DNSSEC_LOCAL_VALIDATION
 		hp = gethostbyname(hn);
+#else
+                hp = val_gethostbyname(NULL, hn, &val_status);
+#endif
 		if (t > 0)
 			(void) alarm(0);
 		(void) NcSignal(SIGPIPE, osigpipe);
 		(void) NcSignal(SIGINT, osigint);
 		(void) NcSignal(SIGALRM, osigalrm);
+#ifdef DNSSEC_LOCAL_VALIDATION
+                /*
+                 * It would be nice to pass a little more information back,
+                 * but that would mean an API change to MyGetHostByName.
+                 */
+                if ((hp != NULL) && ! val_istrusted(val_status)) {
+                    *dst = '\0';
+                    return (-2);
+                }
+#endif
 		if (hp != NULL) {
 			InetNtoA(dst, ((struct in_addr **) hp->h_addr_list)[0], dsize);
 			return (0);
@@ -859,7 +899,6 @@ strncoll(const char *a, const char *b, size_t n)
 		(void) strncpy(sa, a, n);
 		(void) strncpy(sb, b, n);
 		rc = (strcoll(sa, sb));
-		return (rc);
 	} else {
 		char *ma, *mb;
 
@@ -877,8 +916,8 @@ strncoll(const char *a, const char *b, size_t n)
 		rc = (strcoll(ma, mb));
 		free(ma);
 		free(mb);
-		return (rc);
 	}
+	return (rc);
 }	/* strncoll */
 #endif
 
