@@ -1,6 +1,6 @@
 /* ncftpput.c
  *
- * Copyright (c) 1996-2004 Mike Gleason, NcFTP Software.
+ * Copyright (c) 1996-2005 Mike Gleason, NcFTP Software.
  * All rights reserved.
  *
  * A simple, non-interactive utility to send files to a remote FTP server.
@@ -14,14 +14,16 @@
 
 #if (defined(WIN32) || defined(_WINDOWS)) && !defined(__CYGWIN__)
 #	include "..\ncftp\util.h"
+#	include "..\ncftp\bookmark.h"
 #	include "..\ncftp\spool.h"
 #	include "..\ncftp\pref.h"
-#	include "..\ncftp\getline.h"
+#	include "..\ncftp\gl_getline.h"
 #else
 #	include "../ncftp/util.h"
+#	include "../ncftp/bookmark.h"
 #	include "../ncftp/spool.h"
 #	include "../ncftp/pref.h"
-#	include "../ncftp/getline.h"
+#	include "../ncftp/gl_getline.h"
 #endif
 
 #include "gpshare.h"
@@ -37,6 +39,7 @@ extern unsigned int gFirewallPort;
 extern char gFirewallExceptionList[256];
 extern int gFwDataPortMode;
 extern const char gOS[], gVersion[];
+extern Bookmark gBm;
 
 static void
 Usage(void)
@@ -46,9 +49,11 @@ Usage(void)
 	fp = OpenPager();
 	(void) fprintf(fp, "NcFTPPut %.5s\n\n", gVersion + 11);
 	(void) fprintf(fp, "Usages:\n");
-	(void) fprintf(fp, "  ncftpput [flags] remote-host remote-dir local-files...   (mode 1)\n");
+	(void) fprintf(fp, "  ncftpput [flags] remote-host remote-dir local-files...   (mode 1a)\n");
+	(void) fprintf(fp, "  ncftpput [flags] bookmark-name remote-dir local-files... (mode 1b)\n");
 	(void) fprintf(fp, "  ncftpput -f login.cfg [flags] remote-dir local-files...  (mode 2)\n");
-	(void) fprintf(fp, "  ncftpput -c remote-host remote-path-name < stdin  (mode 3)\n");
+	(void) fprintf(fp, "  ncftpput -c remote-host remote-path-name < stdin         (mode 3)\n");
+	(void) fprintf(fp, "  ncftpput -C remote-host local-path-name remote-path-name (mode 4)\n");
 	(void) fprintf(fp, "\nFlags:\n\
   -u XX  Use username XX instead of anonymous.\n\
   -p XX  Use password XX with the username.\n\
@@ -62,9 +67,12 @@ Usage(void)
   -a     Use ASCII transfer type instead of binary.\n\
   -m     Attempt to mkdir the dstdir before copying.\n\
   -v/-V  Do (do not) use progress meters.\n\
-  -f XX  Read the file XX for host, user, and password information.\n");
+  -f XX  Read the file XX for host, user, and password information.\n\
+         If file XX does not exist, check for bookmark XX in $HOME/.ncftp/bookmarks.\n");
 	(void) fprintf(fp, "\
-  -c     Use stdin as input file to write on remote host.\n\
+  -h XX  Connect to host XX.  Useful for overriding host in -f config.file.\n\
+  -c     Read locally from stdin and write remotely to specified pathname.\n\
+  -C     Similar to -c, except a local pathname is specified.\n\
   -A     Append to remote files instead of overwriting them.\n\
   -z/-Z  Do (do not) try to resume uploads (default: -Z).\n\
   -T XX  Upload into temporary files prefixed by XX.\n");
@@ -75,6 +83,7 @@ Usage(void)
   -bb    Same as \"-b\" but queue only (do not run \"ncftpbatch\").\n\
   -E     Use regular (PORT) data connections.\n\
   -F     Use passive (PASV) data connections (default).\n\
+  -I XX  Use IP address XX for local source address.\n\
   -y     Try using \"SITE UTIME\" to preserve timestamps on remote host.\n");
 	(void) fprintf(fp, "\
   -B XX  Try setting the SO_SNDBUF size to XX.\n\
@@ -173,6 +182,8 @@ main(int argc, char **argv)
 	char *Umask = NULL;
 	char *dstdir = NULL;
 	char *dstfile = NULL;
+	const char *preferredLocalAddrStr = NULL;
+	char *lfile = NULL;
 	char **files = (char **) 0;
 	int progmeters;
 	int usingcfg = 0;
@@ -182,10 +193,12 @@ main(int argc, char **argv)
 	int batchmode = 0;
 	int spooled = 0;
 	int i;
+	int ascii = 0;
 	char *ufilep;
 	const char *udirp;
 	char ufile[256];
 	char precmd[320], postcmd[320], perfilecmd[320];
+	int configLoaded = kConfigNotLoaded, configLoaded2 = kConfigNotLoaded;
 	GetoptInfo opt;
 
 	InitWinsock();
@@ -225,7 +238,8 @@ main(int argc, char **argv)
 	perfilecmd[0] = '\0';
 
 	GetoptReset(&opt);
-	while ((c = Getopt(&opt, argc, argv, "P:u:j:p:e:d:U:t:mar:RvVf:o:AT:S:EFcyZzDbB:W:X:Y:")) > 0) {
+#define kGetoptOptions "P:u:j:p:h:e:d:U:t:mar:RvVf:o:AT:S:EFcCyZzDbB:W:X:Y:I:"
+	while ((c = Getopt(&opt, argc, argv, kGetoptOptions)) > 0) {
 		if (c == 'b') {
 			batchmode++;
 		}
@@ -233,11 +247,11 @@ main(int argc, char **argv)
 
 	if (batchmode > 0) {
 		GetoptReset(&opt);
-		while ((c = Getopt(&opt, argc, argv, "P:u:j:p:e:d:U:t:mar:RvVf:AT:S:EFcyZzDbB:W:X:Y:")) > 0) switch(c) {
+		while ((c = Getopt(&opt, argc, argv, kGetoptOptions)) > 0) switch(c) {
 			case 'v': case 'V': case 'A': case 'B': case 'S':
 			case 'T': case 'd': case 'e': case 'U': case 't':
 			case 'm': case 'r': case 'c': case 'y': case 'z':
-			case 'Z':
+			case 'Z': case 'C':
 				(void) fprintf(stderr, "The \"-%c\" option is not valid when used with conjunction with \"-%c\".\n", c, 'b');
 				exit(kExitUsage);
 				break;
@@ -245,24 +259,13 @@ main(int argc, char **argv)
 	}
 
 	GetoptReset(&opt);
-	while ((c = Getopt(&opt, argc, argv, "P:u:j:p:e:d:U:t:mar:RvVf:o:AT:S:EFcyZzDbB:W:X:Y:")) > 0) switch(c) {
+	while ((c = Getopt(&opt, argc, argv, kGetoptOptions)) > 0) switch(c) {
 		case 'P':
-			fi.port = atoi(opt.arg);	
-			break;
 		case 'u':
-			(void) STRNCPY(fi.user, opt.arg);
-			memset(opt.arg, 0, strlen(fi.user));
-			opt.arg[0] = '?';
-			break;
 		case 'j':
-			(void) STRNCPY(fi.acct, opt.arg);
-			memset(opt.arg, 0, strlen(fi.acct));
-			opt.arg[0] = '?';
-			break;
 		case 'p':
-			(void) STRNCPY(fi.pass, opt.arg);	/* Don't recommend doing this! */
-			memset(opt.arg, 0, strlen(fi.pass));
-			opt.arg[0] = '?';
+		case 'h':
+			/* Use these options later */
 			break;
 		case 'e':
 			if (strcmp(opt.arg, "stdout") == 0)
@@ -271,8 +274,10 @@ main(int argc, char **argv)
 				fi.errLog = stdout;
 			else if (strcmp(opt.arg, "stderr") == 0)
 				fi.errLog = stderr;
-			else
+			else {
 				fi.errLog = fopen(opt.arg, FOPEN_APPEND_TEXT);
+				fi.debugTimestamping = 2;
+			}
 			break;
 		case 'D':
 			/* Require two -D's in case they typo. */
@@ -285,8 +290,10 @@ main(int argc, char **argv)
 				fi.debugLog = stdout;
 			else if (strcmp(opt.arg, "stderr") == 0)
 				fi.debugLog = stderr;
-			else
+			else {
 				fi.debugLog = fopen(opt.arg, FOPEN_APPEND_TEXT);
+				fi.debugTimestamping = 2;
+			}
 			break;
 		case 'U':
 			Umask = opt.arg;
@@ -299,6 +306,7 @@ main(int argc, char **argv)
 			break;
 		case 'a':
 			xtype = kTypeAscii;	/* Use ascii. */
+			ascii++;
 			break;
 		case 'r':
 			SetRedial(&fi, opt.arg);
@@ -313,7 +321,10 @@ main(int argc, char **argv)
 			progmeters = 0;
 			break;
 		case 'f':
-			ReadConfigFile(opt.arg, &fi);
+			if ((configLoaded = ReadConfigFile(opt.arg, &fi)) == kErrorLoadingConfig) {
+				perror(opt.arg);
+				exit(kExitBadConfigFile);
+			}
 			usingcfg = 1;
 			break;
 		case 'o':
@@ -337,6 +348,9 @@ main(int argc, char **argv)
 		case 'c':
 			ftpcat = 1;
 			break;
+		case 'C':
+			ftpcat = 2;
+			break;
 		case 'y':
 			tryUtime = 1;
 			break;
@@ -364,25 +378,57 @@ main(int argc, char **argv)
 			STRNCAT(postcmd, opt.arg);
 			STRNCAT(postcmd, "\n");
 			break;
+		case 'I':
+			if (AddrStrToAddr(opt.arg, &fi.preferredLocalAddr, 21) < 0) {
+				fprintf(stderr, "Bad IP address (\"%s\") used with -I.\n", opt.arg);
+				Usage();
+			}
+			preferredLocalAddrStr = opt.arg;
+			break;
 		default:
 			Usage();
 	}
 
-	if ((wantMkdir != 0) && (ftpcat != 0)) {
-		(void) fprintf(stderr, "The \"-%c\" option is not valid when used with conjunction with \"-%c\".\n", 'm', 'c');
-		exit(kExitUsage);
+	if ((rflag != 0) && (ftpcat != 0)) {
+		fprintf(stderr, "The -R flag is not supported with -c/-C.\n\n");
+		Usage();
 	}
 
+	if ((perfilecmd[0] != '\0') && (rflag != 0))
+		(void) fprintf(stderr, "Warning: your -X command is only applied once per command-line parameter, and not for each file in the directory.\n");
+
 	if (usingcfg != 0) {
 		if (ftpcat == 0) {
 			if (opt.ind > argc - 2)
 				Usage();
 			dstdir = argv[opt.ind + 0];
 			files = argv + opt.ind + 1;
-		} else {
-			if (opt.ind > argc - 1)
+		} else if (ftpcat == 2) {
+			if ((opt.ind + 2) > argc) {
 				Usage();
-			dstfile = argv[opt.ind + 1];
+			} else if ((opt.ind + 2) == argc) {
+				lfile = argv[opt.ind + 0];
+				dstfile = argv[opt.ind + 1];
+			} else	/* if ((opt.ind + 2) < argc) */ {
+				/* host argument at +0 will be overwritten;
+				 * the host needs to be in the config file.
+				 */
+				(void) STRNCPY(fi.host, argv[opt.ind + 0]);
+				lfile = argv[opt.ind + 1];
+				dstfile = argv[opt.ind + 2];
+			}
+		} else /* (ftpcat == 1) */ {
+			if ((opt.ind + 1) > argc) {
+				Usage();
+			} else if ((opt.ind + 1) == argc) {
+				dstfile = argv[opt.ind + 0];
+			} else	/* if ((opt.ind + 1) < argc) */ {
+				/* host argument at +0 will be overwritten;
+				 * the host needs to be in the config file.
+				 */
+				(void) STRNCPY(fi.host, argv[opt.ind + 0]);
+				dstfile = argv[opt.ind + 1];
+			}
 		}
 	} else {
 		if (ftpcat == 0) {
@@ -391,16 +437,60 @@ main(int argc, char **argv)
 			(void) STRNCPY(fi.host, argv[opt.ind]);
 			dstdir = argv[opt.ind + 1];
 			files = argv + opt.ind + 2;
+		} else if (ftpcat == 2) {
+			if (opt.ind > argc - 3)
+				Usage();
+			(void) STRNCPY(fi.host, argv[opt.ind]);
+			lfile = argv[opt.ind + 1];
+			dstfile = argv[opt.ind + 2];
 		} else {
 			if (opt.ind > argc - 2)
 				Usage();
 			(void) STRNCPY(fi.host, argv[opt.ind]);
 			dstfile = argv[opt.ind + 1];
 		}
+		if (strchr(fi.host, '.') == NULL) {
+			/* Check if host argument corresponds to config file or entry from ~/.ncftp/bookmarks. */
+
+			configLoaded2 = ReadConfigFile(fi.host, &fi);
+		}
+	}
+
+	/* In case we loaded a bookmark or config file, we allow the following
+	 * options to override those settings.
+	 */
+	GetoptReset(&opt);
+	while ((c = Getopt(&opt, argc, argv, kGetoptOptions)) > 0) switch(c) {
+		case 'P':
+			fi.port = atoi(opt.arg);	
+			break;
+		case 'u':
+			(void) STRNCPY(fi.user, opt.arg);
+			memset(opt.arg, 0, strlen(fi.user));
+			opt.arg[0] = '?';
+			break;
+		case 'j':
+			(void) STRNCPY(fi.acct, opt.arg);
+			memset(opt.arg, 0, strlen(fi.acct));
+			opt.arg[0] = '?';
+			break;
+		case 'p':
+			(void) STRNCPY(fi.pass, opt.arg);	/* Don't recommend doing this! */
+			fi.leavePass = 1;
+			if (fi.pass[0] == '\0')
+				fi.passIsEmpty = 1;
+			memset(opt.arg, 0, strlen(fi.pass));
+			opt.arg[0] = '?';
+			break;
+		case 'h':
+			(void) STRNCPY(fi.host, opt.arg);
+			memset(opt.arg, 0, strlen(fi.user));
+			opt.arg[0] = '?';
+			break;
 	}
 
 	if (strcmp(fi.user, "anonymous") && strcmp(fi.user, "ftp")) {
-		if (fi.pass[0] == '\0') {
+		if ((fi.pass[0] == '\0') && (fi.passIsEmpty == 0)) {
 			(void) gl_getpass("Password: ", fi.pass, sizeof(fi.pass));
 		}
 	}
@@ -411,6 +501,12 @@ main(int argc, char **argv)
 		fi.hasSITE_UTIME = 0;
 	if (nD >= 2)
 		deleteflag = kDeleteYes;
+	if (ascii > 1) {
+		/* This is used internally for testing.
+		 * Valid values as of this release are -1, 0, 1.
+		 */
+		fi.asciiTranslationMode = ascii - 3;
+	}
 
 	if (MayUseFirewall(fi.host, gFirewallType, gFirewallExceptionList) != 0) {
 		fi.firewallType = gFirewallType; 
@@ -462,10 +558,11 @@ main(int argc, char **argv)
 				(time_t) 0,	/* when: now */
 				0,
 				fi.manualOverrideFeatures,
+				preferredLocalAddrStr,
 				0
 			);
 			if ((result == 0) && (batchmode < 3)) {
-				fprintf(stdout, "  + Spooled; sending remotely as %s/%s.\n", dstdir, ufilep);
+				(void) fprintf(stdout, "  + Spooled; sending remotely as %s/%s.\n", dstdir, ufilep);
 				spooled++;
 			}
 		}
@@ -499,10 +596,27 @@ main(int argc, char **argv)
 
 	(void) AdditionalCmd(&fi, precmd, NULL);
 
+	if ((configLoaded2 == kLoadedFromBookmark) && (strcmp(gBm.dir, "/") != 0) && (strcmp(gBm.dir, ".") != 0)) {
+		if ((dstdir != NULL) && ((dstdir[0] == '/') || (dstdir[0] == '\\'))) {
+			/* Don't bother if we are just going to change to an absolute path next */
+			result = kNoErr;
+		} else if ((result = FTPChdir(&fi, gBm.dir)) < 0) {
+			FTPPerror(&fi, result, kErrCWDFailed, "Warning: Could not change to bookmark's remote directory of", gBm.dir);
+		}
+
+	}
+
 	if (dstdir != NULL) {
 		es = kExitChdirTimedOut;
 		if (wantMkdir != 0) {
 			result = FTPChdir3(&fi, dstdir, NULL, 0, kChdirFullPath|kChdirOneSubdirAtATime|kChdirAndMkdir);
+			if (result == kErrMKDFailed) {
+				FTPPerror(&fi, result, kErrMKDFailed, "ncftpput: Could not create directory", dstdir);
+				(void) FTPCloseHost(&fi);
+				es = kExitMkdirFailed;
+				DisposeWinsock();
+				exit((int) es);
+			}
 		} else {
 			result = FTPChdir3(&fi, dstdir, NULL, 0, kChdirFullPath|kChdirOneSubdirAtATime);
 		}
@@ -524,11 +638,29 @@ main(int argc, char **argv)
 				es = kExitXferFailed;
 			else
 				es = kExitSuccess;
+		} else if (ftpcat == 2) {
+			result = kNoErr;
+			if (wantMkdir)
+				result = FTPMkParentDir(&fi, dstfile, 1, NULL);
+			if (result != kNoErr) {
+				es = kExitMkdirFailed;
+			} else if ((result = FTPPutOneFile3(&fi, lfile, dstfile, xtype, -1, appendflag, tmppfx, tmpsfx, resumeflag, deleteflag, kNoFTPConfirmResumeUploadProc, 0)) < 0) {
+				FTPPerror(&fi, result, kErrCouldNotStartDataTransfer, "ncftpput", dstfile);
+				es = kExitXferFailed;
+			} else {
+				es = kExitSuccess;
+				(void) AdditionalCmd(&fi, perfilecmd, argv[opt.ind + 1]);
+			}
 		} else {
 			fi.progress = (FTPProgressMeterProc) 0;
-			if (FTPPutOneFile2(&fi, NULL, dstfile, xtype, STDIN_FILENO, appendflag, tmppfx, tmpsfx) < 0)
+			result = kNoErr;
+			if (wantMkdir)
+				result = FTPMkParentDir(&fi, dstfile, 1, NULL);
+			if (result != kNoErr) {
+				es = kExitMkdirFailed;
+			} else if (FTPPutOneFile3(&fi, NULL, dstfile, xtype, STDIN_FILENO, appendflag, tmppfx, tmpsfx, resumeflag, deleteflag, kNoFTPConfirmResumeUploadProc, 0) < 0) {
 				es = kExitXferFailed;
-			else {
+			} else {
 				es = kExitSuccess;
 				(void) AdditionalCmd(&fi, perfilecmd, argv[opt.ind + 1]);
 			}
