@@ -3,6 +3,8 @@
 #	pragma hdrstop
 #endif
 
+#define _CRT_SECURE_NO_WARNINGS 1
+
 #ifndef STRNCPY
 #	define STRNCPY(a,b) strncpy(a, b, sizeof(a)); a[sizeof(a) - 1] = '\0'
 #	define Strncpy(a,b,s) strncpy(a, b, s); a[s - 1] = '\0' 
@@ -16,7 +18,7 @@
 	}
 #endif
 
-#if (((defined(MACOSX)) && (MACSOX < 10300)) || (defined(AIX) && (AIX < 430)) || (defined(DIGITAL_UNIX)) || (defined(SOLARIS)) || (defined(SCO)) || (defined(HPUX)))
+#if (((defined(MACOSX)) && (MACOSX < 10300)) || (defined(AIX) && (AIX < 430)) || (defined(DIGITAL_UNIX)) || (defined(SOLARIS)) || (defined(SCO)) || (defined(HPUX)))
 extern int getdomainname(char *name, gethostname_size_t namelen);
 #endif
 
@@ -24,14 +26,38 @@ extern int getdomainname(char *name, gethostname_size_t namelen);
 int
 GetHostByName(struct hostent *const hp, const char *const name, char *const hpbuf, size_t hpbufsize)
 {
-#if defined(HAVE_GETHOSTBYNAME_R) && (defined(SOLARIS) || defined(IRIX) || defined(BSDOS))
+#if defined(DNSSEC_LOCAL_VALIDATION)
+	char *usehpbuf;
+	struct hostent *h;
+	int my_h_errno, rc;
+        val_status_t val_status;
+
+	usehpbuf = hpbuf;
+	forever {
+		errno = 0;
+		my_h_errno = 0;
+		h = NULL;
+		memset(usehpbuf, 0, hpbufsize);
+		rc = val_gethostbyname2_r(NULL, name, AF_INET, hp, usehpbuf,
+                                          hpbufsize, &h, &my_h_errno,
+                                          &val_status);
+		if ((rc == 0) && (h != NULL)) {
+                    if (!val_istrusted(val_status))
+			return (-2);
+                    return (0);
+                }
+		if ((rc == 0) && (my_h_errno != 0))
+			errno = ENOENT;
+		break;
+	}
+#elif defined(HAVE_GETHOSTBYNAME_R) && (defined(SOLARIS) || defined(IRIX) || defined(BSDOS))
 	struct hostent *h;
 	int h_errno_unused = 0;
 	memset(hpbuf, 0, hpbufsize);
 	h = gethostbyname_r(name, hp, hpbuf, hpbufsize, &h_errno_unused);
 	if (h != NULL)
 		return (0);
-#elif defined(HAVE_GETHOSTBYNAME2_R) && defined(LINUX)
+#elif defined(HAVE_GETHOSTBYNAME2_R) && defined(LINUX) && defined(HAVE_ALLOCA)
 	char *usehpbuf;
 	struct hostent *h;
 	int my_h_errno, rc;
@@ -52,12 +78,13 @@ GetHostByName(struct hostent *const hp, const char *const name, char *const hpbu
 				errno = ENOMEM;
 				return (-1);
 			}
+			continue;
 		}
 		if ((rc == 0) && (my_h_errno != 0))
 			errno = ENOENT;
 		break;
 	}
-#elif defined(HAVE_GETHOSTBYNAME_R) && defined(LINUX)
+#elif defined(HAVE_GETHOSTBYNAME_R) && defined(LINUX) && defined(HAVE_ALLOCA)
 	char *usehpbuf;
 	struct hostent *h;
 	int my_h_errno, rc;
@@ -78,6 +105,7 @@ GetHostByName(struct hostent *const hp, const char *const name, char *const hpbu
 				errno = ENOMEM;
 				return (-1);
 			}
+			continue;
 		}
 		if ((rc == 0) && (my_h_errno != 0))
 			errno = ENOENT;
@@ -110,14 +138,28 @@ GetHostByName(struct hostent *const hp, const char *const name, char *const hpbu
 int
 GetHostByAddr(struct hostent *const hp, void *addr, int asize, int atype, char *const hpbuf, size_t hpbufsize)
 {
-#if defined(HAVE_GETHOSTBYADDR_R) && (defined(SOLARIS) || defined(IRIX) || defined(BSDOS))
+#if defined(DNSSEC_LOCAL_VALIDATION)
+	struct hostent *h;
+	int h_errno_unused = 0, rc;
+	val_status_t val_status;
+
+	memset(hpbuf, 0, hpbufsize);
+	rc = val_gethostbyaddr_r(NULL, addr, asize, atype,
+                                 hp, hpbuf, hpbufsize, &h, &h_errno_unused,
+                                 &val_status);
+	if ((rc == 0) && (h != NULL)) {
+		if (!val_istrusted(val_status))
+			return (-2);
+		return (0);
+	}
+#elif defined(HAVE_GETHOSTBYADDR_R) && (defined(SOLARIS) || defined(IRIX) || defined(BSDOS))
 	struct hostent *h;
 	int h_errno_unused = 0;
 	memset(hpbuf, 0, hpbufsize);
 	h = gethostbyaddr_r((gethost_addrptr_t) addr, asize, atype, hp, hpbuf, hpbufsize, &h_errno_unused);
 	if (h != NULL)
 		return (0);
-#elif defined(HAVE_GETHOSTBYADDR_R) && defined(LINUX)
+#elif defined(HAVE_GETHOSTBYADDR_R) && defined(LINUX) && defined(HAVE_ALLOCA)
 	char *usehpbuf;
 	struct hostent *h;
 	int my_h_errno, rc;
@@ -138,6 +180,7 @@ GetHostByAddr(struct hostent *const hp, void *addr, int asize, int atype, char *
 				errno = ENOMEM;
 				return (-1);
 			}
+			continue;
 		}
 		if ((rc == 0) && (my_h_errno != 0))
 			errno = ENOENT;
@@ -177,7 +220,7 @@ int
 GetHostEntry(struct hostent *const hp, const char *const host, struct in_addr *const ip_address, char *const hpbuf, size_t hpbufsize)
 {
 	struct in_addr ip;
-	int rc = -1;
+	int rc;
 	
 	/* See if the host was given in the dotted IP format, like "36.44.0.2."
 	 * If it was, inet_addr will convert that to a 32-bit binary value;
@@ -185,7 +228,8 @@ GetHostEntry(struct hostent *const hp, const char *const host, struct in_addr *c
 	 */
 	ip.s_addr = inet_addr(host);
 	if (ip.s_addr != INADDR_NONE) {
-		if (GetHostByAddr(hp, (char *) &ip, (int) sizeof(ip), AF_INET, hpbuf, hpbufsize) == 0) {
+		rc = GetHostByAddr(hp, (char *) &ip, (int) sizeof(ip), AF_INET, hpbuf, hpbufsize);
+		if (rc == 0) {
 			rc = 0;
 			if (ip_address != NULL)
 				(void) memcpy(&ip_address->s_addr, hp->h_addr_list[0], (size_t) hp->h_length);
@@ -196,12 +240,19 @@ GetHostEntry(struct hostent *const hp, const char *const host, struct in_addr *c
 		/* No IP address, so it must be a hostname, like ftp.wustl.edu. */
 		if (ip_address != NULL)
 			ip_address->s_addr = INADDR_NONE;
-		if (GetHostByName(hp, host, hpbuf, hpbufsize) == 0) {
-			rc = 0;
+		rc = GetHostByName(hp, host, hpbuf, hpbufsize);
+		if (rc == 0) {
 			if (ip_address != NULL)
 				(void) memcpy(&ip_address->s_addr, hp->h_addr_list[0], (size_t) hp->h_length);
 		}
 	}
+#if defined(DNSSEC_LOCAL_VALIDATION)
+	if ((rc < 0) && (rc != -2))
+		rc = -1;
+#else
+	if ((rc < 0) && (rc != -1))
+		rc = -1;
+#endif
 	return (rc);
 }	/* GetHostEntry */
 
